@@ -3,12 +3,17 @@ from functools import wraps
 import importlib.resources
 import inspect
 import json
+from os import PathLike
 from typing import Callable, Optional
 from dataclasses import dataclass, field
-from numpy import int64
+import dask
+import dask.dataframe
+import dask.array
+import numpy
 
 import pandas
 import requests
+import outliers
 import scipy.stats
 
 from datapact.util.github import get_github_url
@@ -58,6 +63,11 @@ class Expectation:  # pylint: disable=too-many-instance-attributes
             else None,
         }
 
+    def __repr__(self) -> str:
+        if self.success:
+            return f"Pass({self.name})"
+        return f"Fail({self.name}: {self.message})"
+
     def _repr_markdown_(self):
         if self.parent is not None:
             return self.parent._repr_markdown_()  # pylint: disable=protected-access
@@ -68,6 +78,9 @@ class Expectation:  # pylint: disable=too-many-instance-attributes
             return self.parent._repr_html_()  # pylint: disable=protected-access
         return None
 
+    def __bool__(self):
+        return self.success
+
 
 @dataclass
 class SeriesResult:
@@ -112,11 +125,17 @@ def compute(value):
     return value
 
 
-class SeriesTest:
-    """
-    wraps a column
-    """
+def check_range(min_value: float, max_value: float, value: float, name):
+    result = {"value": value}
+    if value < min_value or value > max_value:
+        return Expectation.Fail(
+            f"{name} is not in range [{min_value}, {max_value}], found {value}",
+            **result,
+        )
+    return Expectation.Pass()
 
+
+class SeriesTest:
     def __init__(self, parent: "DataframeTest", series: pandas.Series):
         self.parent = parent
         self.series = series
@@ -133,7 +152,10 @@ class SeriesTest:
         unit: Optional[str] = None,
     ):
         """
-        demo docstring
+        Adds human-readable information
+        to the column report.
+
+        >>> dp.SepalWidth.describe(description="TODO: put some sepal knowledge in here", unit="cm")
         """
         if title is not None:
             self.title = title
@@ -147,7 +169,7 @@ class SeriesTest:
 def coerce_to_json(arg):
     if inspect.isfunction(arg):
         return arg.__name__
-    if isinstance(arg, int64):
+    if isinstance(arg, numpy.int64):
         return str(arg)
     return arg
 
@@ -216,6 +238,11 @@ class Asserter:
             >>> dp.salary.should.be_normal(alpha=0.1)
         """
 
+        if not pandas.api.types.is_numeric_dtype(self.series):
+            return Expectation.Fail(
+                "not numeric. cannot perform normaltest.",
+            )
+
         stat, p = scipy.stats.normaltest(self.series)
         bins = self.bins()
 
@@ -254,8 +281,8 @@ class Asserter:
 
         if extends_left and extends_right:
             return Expectation.Fail(
-                f"out of range. expected to be in ({minimum}, {maximum}) \n"
-                + f"but found ({found_min}, {found_max}).",
+                f"expected values to be in ({minimum}, {maximum}), "
+                + f"but found ({found_min}, {found_max})",
                 failed_sample_indices=[
                     compute(self.series.idxmin()),
                     compute(self.series.idxmax()),
@@ -308,11 +335,11 @@ class Asserter:
             >>> dp.debt.should.be_negative()
         """
 
-        found_max = self.series.max()
+        found_max = compute(self.series.max())
 
         result = {"maximum": found_max}
 
-        if found_max < 0:
+        if found_max > 0:
             return Expectation.Fail(
                 f"positive value found: max is {found_max}",
                 failed_sample_indices=[compute(self.series.idxmax())],
@@ -321,25 +348,6 @@ class Asserter:
 
         return Expectation.Pass(**result)
 
-    @expectation
-    def not_be_null(self):
-        """
-        checks if all values are non-null.
-
-        Examples:
-            >>> dp.user_id.must.not_be_null()
-        """
-
-        null_index = self.series.isnull().first_valid_index()
-
-        if null_index:
-            return Expectation.Fail(
-                "found null values",
-                failed_sample_indices=[null_index],
-            )
-
-        return Expectation.Pass()
-
     @expectation
     def not_be_na(self):
         """
@@ -349,10 +357,13 @@ class Asserter:
             >>> dp.user_id.must.not_be_na()
         """
 
-        na_index = self.series.isna().first_valid_index()
+        isna = self.series.isna()
 
-        if na_index:
-            return Expectation.Fail("found na values", failed_sample_indices=[na_index])
+        if compute(isna.any()):
+            failed_sample_indices = compute(self.series[isna].index).to_list()[:5]
+            return Expectation.Fail(
+                "found na values", failed_sample_indices=failed_sample_indices
+            )
 
         return Expectation.Pass()
 
@@ -397,6 +408,17 @@ class Asserter:
             >>> dp.day.must.be_date()
         """
 
+        if self.series.dtype == numpy.dtype("datetime64[ns]"):
+            seconds_not_zero = self.series[
+                self.series.dt.time != datetime.time.fromisoformat("00:00")
+            ]
+            if len(seconds_not_zero) > 0:
+                return Expectation.Fail(
+                    "found datetimes with non-zero seconds",
+                    failed_sample_indices=seconds_not_zero.index,
+                )
+            return Expectation.Pass()
+
         rejected = compute(
             self.series[-self.series.str.match(r"\d{4}-\d{2}-\d{2}")].index
         ).to_list()
@@ -405,7 +427,6 @@ class Asserter:
             return Expectation.Fail(
                 "found non-date values", failed_sample_indices=rejected[:5]
             )
-
         return Expectation.Pass()
 
     @expectation
@@ -417,13 +438,18 @@ class Asserter:
             >>> dp.timestamp.must.be_datetime()
         """
 
-        rejected = compute(
-            self.series[-pandas.to_datetime(self.series, errors="coerce").isna()].index
-        ).to_list()
+        to_datetime = (
+            pandas.to_datetime
+            if isinstance(self.series, pandas.Series)
+            else dask.dataframe.to_datetime
+        )
 
-        if len(rejected) > 0:
+        invalid_datetimes = to_datetime(self.series, errors="coerce").isna()
+        invalid_rows = compute(self.series[invalid_datetimes].index).to_list()
+
+        if len(invalid_rows) > 0:
             return Expectation.Fail(
-                "found non-datetime values", failed_sample_indices=rejected[:5]
+                "found non-datetime values", failed_sample_indices=invalid_rows[:5]
             )
 
         return Expectation.Pass()
@@ -452,7 +478,7 @@ class Asserter:
                 **result,
             )
 
-        max_timestamp = datetime.datetime(2100, 0, 0).timestamp()
+        max_timestamp = datetime.datetime(2100, 1, 1).timestamp()
         if found_max > max_timestamp:
             return Expectation.Fail(
                 f"found {found_max}, which is after the year 2100",
@@ -462,6 +488,172 @@ class Asserter:
 
         return Expectation.Pass(**result)
 
+    @expectation
+    def match_sample(self, sample, alpha=0.05):
+        """
+        checks if series is from same distribution
+        as sample using a kolmogorov-smirnoff-test.
+
+        Args:
+            sample : list-like
+                sample to compare to
+
+        Examples:
+            >>> dp.age.should.match_sample(reference_sample)
+        """
+
+        stat, p = scipy.stats.ks_2samp(self.series, sample)
+        result = {
+            "stat": stat,
+            "p": p,
+        }
+
+        if p < alpha:
+            return Expectation.Fail("kolmogorov-smirnoff-test rejected", **result)
+
+        return Expectation.Pass(**result)
+
+    def _match_cdf(self, cdf: Callable, args, N=20, alpha=0.05):
+        """
+        checks if series is from distribution as given by cdf using a kolmogorov-smirnoff-test.
+
+        Args:
+            cdf : Callable
+                Used to calculate the cdf.
+            args :
+                Distribution parameters, given to cdf.
+
+        Examples:
+            >>> dp.wins.should.match_sample(scipy.stats.binom)
+        """
+
+        stat, p = scipy.stats.kstest(self.series, cdf, args, N)
+        result = {"stat": stat, "p": p, "bins": self.bins()}
+
+        if p < alpha:
+            return Expectation.Fail("kolmogorov-smirnoff-test rejected", **result)
+
+        return Expectation.Pass(**result)
+
+    @expectation
+    def match_cdf(self, cdf: Callable, args, N=20, alpha=0.05):
+        """
+        checks if series is from distribution as given by cdf using a kolmogorov-smirnoff-test.
+
+        Args:
+            cdf : Callable
+                Used to calculate the cdf.
+            args :
+                Distribution parameters, given to cdf.
+
+        Examples:
+            >>> dp.wins.should.match_sample(scipy.stats.binom)
+        """
+
+        return self._match_cdf(cdf, args, N=N, alpha=alpha)
+
+    @expectation
+    def be_binomial_distributed(self, n, p=0.5, N=20, alpha=0.05):
+        """
+        checks if series is binomial distributed using a kolmogorov-smirnoff-test.
+
+        Args:
+            n : number
+                number of draws
+            p :
+                probability of success
+
+        Examples:
+            >>> dp.heads.should.be_binomial_distributed()
+        """
+
+        return self._match_cdf(
+            scipy.stats.distributions.binom.cdf, (n, p), N=N, alpha=alpha
+        )
+
+    @expectation
+    def be_poisson_distributed(self, l, N=20, alpha=0.05):
+        """
+        checks if series is poisson distributed using a kolmogorov-smirnoff-test.
+
+        Args:
+            l : number
+                lambda for poission distribution
+
+        Examples:
+            >>> dp.new_covid_cases.should.be_poisson_distributed(10)
+        """
+
+        return self._match_cdf(
+            scipy.stats.distributions.poisson.cdf, [l], N=N, alpha=alpha
+        )
+
+    @expectation
+    def have_average_between(self, minimum: float, maximum: float):
+        """
+        checks if average is between min and max.
+
+        Examples:
+            >>> dp.size.should.have_average_between(4, 5)
+        """
+
+        return check_range(minimum, maximum, compute(self.series.mean()), "average")
+
+    @expectation
+    def have_variance_between(self, minimum: float, maximum: float):
+        """
+        checks if average is in given range.
+
+        Examples:
+            >>> dp.size.should.have_variance_between(150, 170)
+        """
+
+        return check_range(minimum, maximum, compute(self.series.var()), "variance")
+
+    @expectation
+    def have_median_between(self, minimum: float, maximum: float):
+        """
+        checks if median is in given range.
+
+        Examples:
+            >>> dp.size.should.have_median_between(150, 170)
+        """
+
+        if not isinstance(self.series, pandas.Series):
+            raise NotImplementedError()
+
+        return check_range(minimum, maximum, self.series.median(), "median")
+
+    @expectation
+    def have_percentile_between(self, p: float, minimum: float, maximum: float):
+        """
+        checks if percentile / quantile is in given range.
+
+        Examples:
+            >>> dp.size.should.have_percentile_between(.95, 10, 20)
+        """
+
+        return check_range(
+            minimum, maximum, compute(self.series.quantile(p)), "percentile"
+        )
+
+    @expectation
+    def have_no_outliers(self, alpha=0.05):
+        """
+        verifies that series has no outliers using Grubbs test.
+
+        Examples:
+            >>> dp.size.should.have_no_outliers()
+        """
+
+        indices = outliers.smirnov_grubbs.two_sided_test_indices(
+            compute(self.series), alpha=alpha
+        )
+        if len(indices) > 0:
+            return Expectation.Fail("found outliers", failed_sample_indices=indices[:5])
+
+        return Expectation.Pass()
+
     @expectation
     def fulfill(self, custom_assertion: Callable[[pandas.Series], Optional[str]]):
         """
@@ -515,7 +707,13 @@ class DataframeTest:
         url: Optional[str] = None,
     ):
         """
-        demo docstring
+        Adds human-readable information
+        to the report.
+
+        >>> dp.describe(
+        >>>     title="Vacancy statistics",
+        >>>     description="data sourced from NYC housing markets"
+        >>> )
         """
         if title is not None:
             self.title = title
@@ -546,6 +744,9 @@ class DataframeTest:
         """
         Collects all tests and returns a DataframeResult object.
         If connected, uploads results.
+
+            >>> dp.day.must.be_datetime()
+            >>> dp.collect() # DataframeResult(...)
         """
 
         result = DataframeResult(self.title, self.description, self.url)
@@ -568,6 +769,10 @@ class DataframeTest:
     def expectations(self) -> "list[Expectation]":
         """
         Returns a list of all executed expectations.
+
+        >>> dp.age.should.be_positive() # Pass(be_positive)
+        >>> dp.day.must.be_datetime() # Fail(be_datetime)
+        >>> dp.failed_critical_expectations() # [ Pass(be_positive), Fail(be_datetime) ]
         """
         result = []
         for s in self.collect().series:
@@ -577,37 +782,58 @@ class DataframeTest:
     def failed_expectations(self) -> "list[Expectation]":
         """
         Returns a list of all failed expectations.
+
+        >>> dp.age.should.be_negative() # Fail(be_negative)
+        >>> dp.day.must.be_datetime() # Fail(be_datetime)
+        >>> dp.failed_critical_expectations() # [ Fail(be_negative), Fail(be_datetime) ]
         """
         return [e for e in self.expectations() if not e.success]
 
     def failed_critical_expectations(self) -> "list[Expectation]":
         """
         Returns a list of all failed critical expectations.
+
+        >>> dp.age.should.be_negative() # Fail(be_negative)
+        >>> dp.day.must.be_datetime() # Fail(be_datetime)
+        >>> dp.failed_critical_expectations() # [ Fail(be_datetime) ]
         """
         return [e for e in self.failed_expectations() if e.critical]
 
     def is_failure(self) -> bool:
         """
         Returns True if one of the expectations failed.
+
+        >>> dp.day.should.be_datetime() # Fail(be_datetime)
+        >>> dp.is_failure() # True
         """
         return len(self.failed_expectations()) > 0
 
     def is_critical_failure(self) -> bool:
         """
         Returns True if a critical expectation failed.
+
+        >>> dp.day.should.be_datetime() # Fail(be_datetime)
+        >>> dp.is_critical_failure() # False
+        >>> dp.day.must.be_datetime() # Fail(be_datetime)
+        >>> dp.is_critical_failure() # True
         """
         return len(self.failed_critical_expectations()) > 0
 
     def check(self):
         """
         Raises an exception if a critical expectation failed.
+
+        >>> dp.day.must.be_datetime() # Fail(be_datetime)
+        >>> dp.check() # 💥
         """
         if self.is_critical_failure():
             raise Exception("critical expectation failed")
 
     def to_html(self):
         """
-        see https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
+        Returns a human-readable test result as HTML.
+
+        >>> dp.to_html() # "<script>..."
         """
         result = self.collect()
 
@@ -628,6 +854,16 @@ class DataframeTest:
 
         return html
 
+    def write_html(self, to: PathLike):
+        """
+        Writes the test result to a file.
+
+        >>> dp.write_html("test.html")
+        """
+
+        with open(to, "w", encoding="utf8") as f:
+            f.write(self.to_html())
+
     def to_markdown(self):
         result = self.collect()
 
@@ -657,5 +893,11 @@ class DataframeTest:
         return self.to_html()
 
 
-def test(dataframe: pandas.DataFrame):
+def test(dataframe: pandas.DataFrame) -> DataframeTest:
+    """
+    Creates a datapact test object.
+
+    >>> df = pandas.read_csv(...)
+    >>> dp = datapact.test(df)
+    """
     return DataframeTest(dataframe)
